﻿namespace Rp1210Example
{
    /// <summary>
    /// Represents a device discovered by calling the DeviceDiscovery IOCTL.
    /// </summary>
    class DiscoveredDevice
    {
        /// <summary>
        /// Initializes a new instance of the DiscoveredDevice class.
        /// </summary>
        /// <param name="name">The name of the discovered device.</param>
        /// <param name="available">Indicates whether the device is available (always false for bluetooth).</param>
        /// <param name="address">The address of the discovered device (IP address for WiFi, MAC address for bluetooth).</param>
        public DiscoveredDevice(string name, bool available, string address)
        {
            Name = name;
            Available = available;
            Address = address;
        }

        /// <summary>
        /// Gets the name of the device.
        /// </summary>
        public string Name
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets a value indicating whether the device is available. This will always return false for bluetooth devices.
        /// </summary>
        public bool Available
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets the address of the device. For a WiFi device, this will return an IP address. For a bluetooth device, this will return a MAC address.
        /// </summary>
        public string Address
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets a value indicating whether the device is a bluetooth device.
        /// </summary>
        public bool IsBluetooth
        {
            get
            {
                if (!string.IsNullOrEmpty(Address))
                {
                    return !Address.Contains(".");
                }
                return false;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the device is a WiFi device.
        /// </summary>
        public bool IsWireless
        {
            get
            {
                if (!string.IsNullOrEmpty(Address))
                {
                    return Address.Contains(".");
                }
                return false;
            }
        }
    }
}