﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using Android.App;
using Android.Content;
using Android.OS;
using Android.Runtime;
using Android.Support.V7.App;
using Android.Widget;
using Java.IO;
using Java.Net;
using Org.Json;

namespace Rp1210Example
{
    [Activity(Label = "@string/app_name", Theme = "@style/AppTheme", MainLauncher = true)]
    public class MainActivity : AppCompatActivity
    {
        const int IOCTL_FW_STATUS = 0x20004;
        const int IOCTL_FW_UPDATE = 0x20005;
        const int IOCTL_PRODUCT_ID = 0x20006;

        const int IOCTL_DEVICE_DISCOVERY = 0x100;
        const int IOCTL_SET_LOG_LEVEL = 0x101;
        

        const int FW_STATUS_QUERY = 0;
        const int FW_STATUS_UPGRADE_IN_PROGRESS = 1;
        const int FW_STATUS_UPGRADE_FINISHED = 2;
        const int FW_STATUS_DEVICE_IN_SYNC = 3;         // The firmware versions match
        const int FW_STATUS_DEVICE_NEWER = 4;           // The firmware is newer on the Miniadapter
        const int FW_STATUS_DEVICE_OLDER = 5;           // A firmware update is available
        const int FW_STATUS_UPDATE_NOT_SUPPORTED = 6;   // Non-Mini adapters not supported
        const int FW_STATUS_UPGRADE_FAILED = 7;
        const int FW_STATUS_UPGRADED_RECONNECT_TO_SYNC = 8; // Not used for Android

        const int RxBufferSize = 2048;
        const int TxBufferSize = 2048;
        const int Rp1210MaxDiscoveryEntryLength = 256;
        private String dataPath = System.Environment.GetFolderPath(System.Environment.SpecialFolder.MyDocuments);
        Task readThreadTask;
        Task connectThreadTask;
        Task getFwVersionThreadTask;
        Task firmwareUpdateThreadTask;
        Task rp1210TestThreadTask;

        Rp1210Api rp1210Api;

        private Button ConnectButton;
        private Button GetVersionsButton;
        private Button DownloadFirmwareButton;
        private Button FwStatusButton;
        private Button FwUpdateButton;
        private Button DisconnectButton;
        private Button Rp1210TestButton;
        private TextView StatusTextView;

        private bool IsConnected = false;
        private short ClientID = -1;
        private string FirmwareLink = "";
        private double FirmwareVersionInCloud = 0;
        SemaphoreSlim GotFwVersionInCloudEvent = new SemaphoreSlim(0, 1);

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
        public struct RP1210FwStatus
        {
            public Int32 status;
            public Int32 file_size;
            public Int32 file_completed;
        }

        struct Rp1210DeviceDiscoveryScan
        {
            public int scanIntervalSec;
            public int maxEntries;
        }

        struct Rp1210SlogLevel
        {
            public IntPtr debugLevel;
            public IntPtr debugFilePath;
            public IntPtr debugFilePathLength;
            public IntPtr debugMode;
            public IntPtr debugFileSize;
        }
        protected override void OnCreate(Bundle savedInstanceState)
        {
            base.OnCreate(savedInstanceState);
            Xamarin.Essentials.Platform.Init(this, savedInstanceState);

            rp1210Api = new Rp1210Api();
            // Set our view from the "main" layout resource
            SetContentView(Resource.Layout.activity_main);

            // Uncompress files that are in Assets/Files
            ExtractAssets();

            StatusTextView = FindViewById<TextView>(Resource.Id.statusTextView);

            // Get our buttons from the layout resource,
            // and attach events to them
            ConnectButton = FindViewById<Button>(Resource.Id.connectButton);
            GetVersionsButton = FindViewById<Button>(Resource.Id.getVersionsButton);
            DownloadFirmwareButton = FindViewById<Button>(Resource.Id.downloadFirmwareButton);
            FwStatusButton = FindViewById<Button>(Resource.Id.firmwareStatusButton);
            FwUpdateButton = FindViewById<Button>(Resource.Id.firmwareUpdateButton);
            DisconnectButton = FindViewById<Button>(Resource.Id.disconnectButton);
            Rp1210TestButton = FindViewById<Button>(Resource.Id.rp1210TestButton);

            ConnectButton.Click += ConnectButton_Click;
            GetVersionsButton.Click += GetVersionsButton_Click;
            DownloadFirmwareButton.Click += DownloadFirmwareButton_Click;
            FwStatusButton.Click += FwStatusButton_Click;
            FwUpdateButton.Click += FwUpdateButton_Click;
            DisconnectButton.Click += DisconnectButton_Click;
            Rp1210TestButton.Click += Rp1210TestButton_Click;
        }

        void ExtractAssets()
        {
            var assets = ApplicationContext.Assets;
            var paths = assets.List("Files");

            foreach (var path in paths)
            {
                // Read the compressed file and extract
                string readPath = Path.Combine("Files", path);
                byte[] buffer = new byte[32000];
                int bytesRead = 0;
                using (var inStream = assets.Open(readPath))
                using (var outStream = ApplicationContext.OpenFileOutput(path, Android.Content.FileCreationMode.Private))
                {
                    do
                    {
                        bytesRead = inStream.Read(buffer, 0, buffer.Length);
                        if (bytesRead > 0)
                        {
                            outStream.Write(buffer, 0, bytesRead);
                        }
                    }
                    while (bytesRead > 0);
                }

                System.Diagnostics.Debug.WriteLine(string.Format("Extracted '{0}' to '{1}'", readPath, path));
            }
        }

        public override void OnRequestPermissionsResult(int requestCode, string[] permissions, [GeneratedEnum] Android.Content.PM.Permission[] grantResults)
        {
            Xamarin.Essentials.Platform.OnRequestPermissionsResult(requestCode, permissions, grantResults);

            base.OnRequestPermissionsResult(requestCode, permissions, grantResults);
        }

        static string GetProtocolString(DiscoveredDevice device)
        {
            return string.Format("{0}:DiscoveryString={1}", "J1708", device.Address);
        }

        private void ConnectButton_Click(object sender, EventArgs e)
        {
            SelectVCI(false);
        }

        private async void GetVersionsButton_Click(object sender, EventArgs e)
        {
            StatusTextView.SetText("Getting versions...", TextView.BufferType.Normal);

            byte[] fpchAPIVersionInfo = new byte[32];
            byte[] fpchDLLVersionInfo = new byte[32];
            byte[] fpchFWVersionInfo = new byte[32];

            if (IsConnected)
            {
                short ret = rp1210Api.ReadDetailedVersion(ClientID, fpchAPIVersionInfo, fpchDLLVersionInfo, fpchFWVersionInfo);

                if (ret == 0)
                {
                    GetFwVersionFromWebService();

                    await GotFwVersionInCloudEvent.WaitAsync(5000);

                    String results;
                    results = "Driver version = " + Encoding.Default.GetString(fpchDLLVersionInfo) + "\n";
                    results += "Firmware version on device = " + Encoding.Default.GetString(fpchFWVersionInfo) + "\n";
                    results += "Firmware version in cloud = " + FirmwareVersionInCloud.ToString();
                    StatusTextView.SetText(results, TextView.BufferType.Normal);
                }
                else
                {
                    StatusTextView.SetText(string.Format("RP1210_ReadDetailedVersion returned {0}", ret), TextView.BufferType.Normal);
                }
            }
            else
            {
                StatusTextView.SetText("Not connected!", TextView.BufferType.Normal);
            }

        }

        private void DownloadFirmwareButton_Click(object sender, EventArgs e)
        {
            if (IsConnected)
            {
                if (FirmwareLink == "")
                {
                    RunOnUiThread(() =>
                    {
                        StatusTextView.SetText("Need to get versions first for firmware link", TextView.BufferType.Normal);
                    });
                }
                else
                {
                    try
                    {
                        int fileNameLen = FirmwareLink.Length - FirmwareLink.LastIndexOf("/") + 1;
                        String fileName = FirmwareLink.Substring(FirmwareLink.LastIndexOf("/") + 1, fileNameLen);

                        String downloadFolder = ApplicationContext.FilesDir.Path;
                        String downloadFwFile = downloadFolder + Java.IO.File.Separator + fileName;

                        System.Net.WebClient webClient = new System.Net.WebClient();
                        webClient.DownloadFileCompleted += new AsyncCompletedEventHandler(FwDownloadCompleted);
                        webClient.DownloadFileAsync(new Uri(FirmwareLink), downloadFwFile);
                    }
                    catch (Exception ex)
                    {
                        System.Diagnostics.Debug.WriteLine("Exception in DownloadFirmwareButton_Click: " + ex.Message);
                    }
                }
            }
            else
            {
                StatusTextView.SetText("Not connected!", TextView.BufferType.Normal);
            }
        }

        private void FwDownloadCompleted(object sender, AsyncCompletedEventArgs e)
        {
            RunOnUiThread(() =>
            {
                if (e.Error != null)
                {
                    StatusTextView.SetText("Download returned ERROR: {0}" + e.Error.Message, TextView.BufferType.Normal);
                }
                else if (!e.Cancelled)
                {
                    StatusTextView.SetText("Download complete", TextView.BufferType.Normal);
                }
            });
        }

        private void FwStatusButton_Click(object sender, EventArgs e)
        {
            RP1210FwStatus fwStatus = new RP1210FwStatus();
            int bufSize = Marshal.SizeOf(fwStatus);
            IntPtr ptrOutput = Marshal.AllocHGlobal(bufSize);

            if (IsConnected)
            {
                short ret = rp1210Api.Ioctl(ClientID, IOCTL_FW_STATUS, IntPtr.Zero, ptrOutput);

                if (ret == 0)
                {
                    String results = "Internal error";
                    fwStatus = (RP1210FwStatus)Marshal.PtrToStructure(ptrOutput, typeof(RP1210FwStatus));

                    switch (fwStatus.status)
                    {
                        case FW_STATUS_QUERY:
                            results = "Query";
                            break;
                        case FW_STATUS_UPGRADE_IN_PROGRESS:
                            results = String.Format("Update in progress: {0}% complete", (int)(100 * ((float)fwStatus.file_completed / (float)fwStatus.file_size)));
                            break;
                        case FW_STATUS_UPGRADE_FINISHED:
                            results = "Update finished";
                            break;
                        case FW_STATUS_DEVICE_IN_SYNC:
                            results = "Device is up to date";
                            break;
                        case FW_STATUS_DEVICE_NEWER:
                            results = "Firmware on device is newer";
                            break;
                        case FW_STATUS_DEVICE_OLDER:
                            results = "Firmware update is available";
                            break;
                        case FW_STATUS_UPDATE_NOT_SUPPORTED:
                            results = "Firmware update is not supported";
                            break;
                        case FW_STATUS_UPGRADE_FAILED:
                            results = "Firmware update FAILED";
                            break;
                    }

                    StatusTextView.SetText(results, TextView.BufferType.Normal);
                }
                else
                {
                    StatusTextView.SetText(string.Format("IOCTL for firmware status returned {0}", ret), TextView.BufferType.Normal);
                }
            }
            else
            {
                StatusTextView.SetText("Not connected!", TextView.BufferType.Normal);
            }
        }

        private void FwUpdateButton_Click(object sender, EventArgs e)
        {
            if (IsConnected)
            {
                StatusTextView.SetText("Updating firmware...\nClick \"Get Firmware Status\" to check progress", TextView.BufferType.Normal);

                GetVersionsButton.Enabled = false;
                DownloadFirmwareButton.Enabled = false;
                DisconnectButton.Enabled = false;
                FwUpdateButton.Enabled = false;
                Rp1210TestButton.Enabled = false;

                firmwareUpdateThreadTask = FirmwareUpdateThreadAsync();
            }
            else
            {
                StatusTextView.SetText("Not connected!", TextView.BufferType.Normal);
            }
        }

        private void DisconnectButton_Click(object sender, EventArgs e)
        {
            if (IsConnected)
            {
                short ret = rp1210Api.ClientDisconnect(ClientID);

                if (ret == 0)
                {
                    StatusTextView.SetText("Disconnect returned success", TextView.BufferType.Normal);
                }
                else
                {
                    StatusTextView.SetText(string.Format("ClientDisconnect returned {0}", ret), TextView.BufferType.Normal);
                }
                IsConnected = false;
            }
            else
            {
                StatusTextView.SetText("Not connected!", TextView.BufferType.Normal);
            }

            ConnectButton.Enabled = true;
        }

        private void Rp1210TestButton_Click(object sender, EventArgs e)
        {
            SelectVCI(true);
        }

        private void SelectVCI(bool IsRunningRp1210Test)
        {
            // Select the type of VCI and set the RP1210 implementation
            ShowAlert(
            "Select type of VCI:",
            new string[] { "Nexiq USB-Link 2", "Nexiq Blue-Link Mini", "Cummins INLINE 7", "Cummins INLINE Mini" },
            (o, args) =>
            {
                switch (args.Which)
                {
                    case 0:
                        rp1210Api.SetImplementation("nuln2r32");
                        DiscoverDevices(IsRunningRp1210Test);
                        break;
                    case 1:
                        rp1210Api.SetImplementation("nblr32");
                        DiscoverDevices(IsRunningRp1210Test);
                        break;
                    case 2:
                        rp1210Api.SetImplementation("cil7r32");
                        DiscoverDevices(IsRunningRp1210Test);
                        break;
                    case 4:
                        rp1210Api.SetImplementation("cimr32");
                        DiscoverDevices(IsRunningRp1210Test);
                        break;
                    default:
                        return;
                }
            });
        }

        void DiscoverDevices(bool IsRunningRp1210Test)
        {
            // Initialize logging after appropriate VCI api is loaded.
            SetupDebugLog();

            // Configure the discovery scan
            Rp1210DeviceDiscoveryScan discovery = new Rp1210DeviceDiscoveryScan()
            {
                maxEntries = 20,
                scanIntervalSec = 3,
            };

            // Marshal the structure to memory
            int sizeDiscovery = Marshal.SizeOf(discovery);
            IntPtr ptrDiscoveryInput = Marshal.AllocHGlobal(sizeDiscovery);
            Marshal.StructureToPtr(discovery, ptrDiscoveryInput, false);

            // Create a buffer for the result
            int sizeResult = sizeof(int) + (discovery.maxEntries * Rp1210MaxDiscoveryEntryLength);
            byte[] outputDiscovery = new byte[sizeResult];
            IntPtr ptrDiscoveryOutput = Marshal.AllocHGlobal(sizeResult);

            // Call the IOCTL that performs the device discovery
            short result = rp1210Api.Ioctl(0, IOCTL_DEVICE_DISCOVERY, ptrDiscoveryInput, ptrDiscoveryOutput);

            Marshal.Copy(ptrDiscoveryOutput, outputDiscovery, 0, (int)sizeResult);

            if (result == 0)
            {
                List<DiscoveredDevice> discoveredDevices = new List<DiscoveredDevice>();

                int numberOfEntries = BitConverter.ToInt32(outputDiscovery, 0);

                byte[] entryData = new byte[Rp1210MaxDiscoveryEntryLength];
                for (int i = 0; i < numberOfEntries; i++)
                {
                    int index = 4 + (i * Rp1210MaxDiscoveryEntryLength);
                    Array.Copy(outputDiscovery, index, entryData, 0, Rp1210MaxDiscoveryEntryLength);
                    string entry = Encoding.UTF8.GetString(entryData);

                    string[] fields = entry.Split(',');
                    string name = fields[0];
                    string serialNum = fields[1];
                    int available = Convert.ToInt32(fields[2]); // 0 = available, 1 = not available
                    string address = fields[3];
                    int endIndex = address.IndexOf('\0');
                    address = address.Substring(0, endIndex);

                    DiscoveredDevice device = new DiscoveredDevice(name, !Convert.ToBoolean(available), address);
                    discoveredDevices.Add(device);
                }

                if (discoveredDevices.Count == 0)
                {
                    ShowAlert("No devices found.");
                }
                else
                {
                    List<string> devices = new List<string>();
                    foreach (var device in discoveredDevices)
                    {
                        devices.Add(string.Format("{0} ({1}), Address = {2}", device.Name, device.IsBluetooth ? "BT" : "WiFi", device.Address));
                    }
                    ShowAlert(
                        "Select device:",
                        devices.ToArray(),
                        (o, args) =>
                        {
                            ConnectDevice(discoveredDevices[args.Which], IsRunningRp1210Test);
                        });
                }
            }
            else
            {
                ShowAlert("Device discovery failed.");
            }
        }

        void ConnectDevice(DiscoveredDevice device, bool IsRunningRp1210Test)
        {
            int client = 0;
            short deviceId = device.IsBluetooth ? (short)2 : (short)3;
            string protocol = GetProtocolString(device);
            short isPacketizingIncomingMsgs = 0;

            if (!IsRunningRp1210Test)
            {
                // Just connect, don't run test
                ConnectButton.Enabled = false;
                StatusTextView.SetText("Connecting...", TextView.BufferType.Normal);
                connectThreadTask = ConnectThreadAsync(client, deviceId, protocol, isPacketizingIncomingMsgs);
            }
            else
            {
                StatusTextView.SetText("Running test...\n(See LogCat for details)", TextView.BufferType.Normal);
                // run test in separate thread
                rp1210TestThreadTask = RP1210TestThreadAsync(client, deviceId, protocol, isPacketizingIncomingMsgs);
            }
        }

        private void SetupDebugLog()
        {
            IntPtr pathValue = (IntPtr)Marshal.StringToHGlobalAnsi(dataPath);

            int pathLength = dataPath.Length + 1;   // add space for null termination.


            Rp1210SlogLevel slogLevel = new Rp1210SlogLevel()
            {
                debugLevel = (IntPtr)4,
                debugFilePath = pathValue,
                debugFilePathLength = (IntPtr)pathLength,
                debugMode = (IntPtr)0,
                debugFileSize = (IntPtr)(16 * 1024 * 1024),
            };

            int sizeLogLevel = Marshal.SizeOf(slogLevel);
            IntPtr ptrSetLogLevelInput = Marshal.AllocHGlobal(sizeLogLevel);
            Marshal.StructureToPtr(slogLevel, ptrSetLogLevelInput, false);

            
            short result = rp1210Api.Ioctl(ClientID, IOCTL_SET_LOG_LEVEL, ptrSetLogLevelInput, IntPtr.Zero);

            if (result != 0)
            {
                System.Diagnostics.Debug.WriteLine("Failed to set log level.");
            }

            Marshal.FreeHGlobal(pathValue);
            Marshal.FreeHGlobal(ptrSetLogLevelInput);
        }

        private void GetFwVersionFromWebService()
        {
            // First get product ID to pass to web service
            Int32 productId = 0;

            if (IsConnected)
            {
                int bufSize = sizeof(uint);
                byte[] outBuf = new byte[sizeof(uint)];
                IntPtr ptrOutput = Marshal.AllocHGlobal(bufSize);

                short ret = rp1210Api.Ioctl(ClientID, IOCTL_PRODUCT_ID, IntPtr.Zero, ptrOutput);

                if (ret == 0)
                {
                    Marshal.Copy(ptrOutput, outBuf, 0, bufSize);
                    productId = BitConverter.ToInt32(outBuf, 0);
                }
                else
                {
                    StatusTextView.SetText(string.Format("IOCTL for product ID returned {0}", ret), TextView.BufferType.Normal);
                }
            }
            else
            {
                StatusTextView.SetText("Not connected!", TextView.BufferType.Normal);
            }

            if (productId == 0)
            {
                return;
            }

            // Request firmware information from web service using product ID
            getFwVersionThreadTask = GetFwVersionThreadAsync(productId);
        }

        void ShowAlert(string message)
        {
            Android.Support.V7.App.AlertDialog.Builder alert = new Android.Support.V7.App.AlertDialog.Builder(this);
            alert.SetMessage(message);
            alert.SetPositiveButton("OK", (d, i) => { });
            alert.Show();
        }

        void ShowAlert(string title, string[] items, EventHandler<DialogClickEventArgs> handler)
        {
            Android.Support.V7.App.AlertDialog.Builder alert = new Android.Support.V7.App.AlertDialog.Builder(this);
            alert.SetTitle(title);
            alert.SetItems(items, handler);
            alert.Show();
        }

        Task ReadThreadAsync(short clientId)
        {
            short result = 0;

            return Task.Run(() =>
            {
                byte[] response = new byte[512];
                result = 1;
                while (result > 0 && result < 128)
                {
                    if ((result = rp1210Api.ReadMessage(clientId, response, (short)response.Length, 1)) > 0)
                    {
                        System.Diagnostics.Debug.WriteLine(string.Format("Read {0} bytes from read thread", result));
                        System.Diagnostics.Debug.WriteLine(hexDump(response, result));
                    }
                }

                System.Diagnostics.Debug.WriteLine("exiting read thread");
            });
        }

        Task ConnectThreadAsync(int client, short deviceId, string protocol, short isPacketizingIncomingMsgs)
        {
            return Task.Run(() =>
            {
                short ret = rp1210Api.ClientConnect(client, deviceId, protocol, TxBufferSize, RxBufferSize, isPacketizingIncomingMsgs);

                if (ret >= 0 && ret < 128)
                {
                    IsConnected = true;
                    ClientID = ret;
                }

                RunOnUiThread(() =>
                {
                    if (IsConnected)
                    {
                        StatusTextView.SetText(string.Format("Connect returned success\nClient ID = {0}", ret), TextView.BufferType.Normal);
                    }
                    else
                    {
                        ConnectButton.Enabled = true;
                        StatusTextView.SetText(string.Format("Connect returned error {0}", ret), TextView.BufferType.Normal);
                    }
                });
            });
        }

        Task GetFwVersionThreadAsync(int productId)
        {
            return Task.Run(() =>
            {
                try
                {
                    HttpURLConnection httpConn = null;

                    String webSvc = "https://prolinkiq.nexiq.com/WebServices/Device.asmx";
                    //String webSvc = "https://prolinkiqbeta.nexiq.com/WebServices/Device.asmx";
                    String urlGetbyProductId = webSvc + "/GetFirmwareVersion?ProductId=" + productId;

                    // send request to get the firmware version that is in the cloud and a link to the fw file
                    URL url = new URL(urlGetbyProductId);
                    httpConn = (HttpURLConnection)url.OpenConnection();
                    httpConn.RequestMethod = "GET";
                    httpConn.ReadTimeout = 10000;
                    httpConn.ConnectTimeout = 15000;
                    httpConn.DoOutput = true;
                    httpConn.Connect();

                    BufferedReader br = null;

                    br = new BufferedReader(new InputStreamReader(url.OpenStream()));

                    StringBuilder sb = new StringBuilder();
                    String line;

                    while ((line = br.ReadLine()) != null)
                    {
                        sb.Append(line + "\n");
                    }

                    br.Close();
                    httpConn.Disconnect();

                    // string is in json format, just a single object
                    String jsonString = sb.ToString();

                    JSONObject jobj = new JSONObject(jsonString);

                    FirmwareLink = jobj.GetString("Link");
                    FirmwareVersionInCloud = Double.Parse(jobj.GetString("Version"));

                    GotFwVersionInCloudEvent.Release();
                }
                catch (Exception e)
                {
                    System.Diagnostics.Debug.WriteLine("Exception in GetFwVersionThreadAsync: " + e.Message);
                }
            });
        }

        Task FirmwareUpdateThreadAsync()
        {
            return Task.Run(() =>
            {
                short ret = rp1210Api.Ioctl(ClientID, IOCTL_FW_UPDATE, IntPtr.Zero, IntPtr.Zero);

                RunOnUiThread(() =>
                {
                    if (ret == 0)
                    {
                        StatusTextView.SetText("Update firmware returned success", TextView.BufferType.Normal);
                    }
                    else
                    {
                        StatusTextView.SetText(String.Format("Update firmware returned error {0}", ret), TextView.BufferType.Normal);
                    }

                    GetVersionsButton.Enabled = true;
                    DownloadFirmwareButton.Enabled = true;
                    DisconnectButton.Enabled = true; ;
                    FwUpdateButton.Enabled = true;
                    Rp1210TestButton.Enabled = true;
                });
            });
        }

        Task RP1210TestThreadAsync(int client, short deviceId, string protocol, short isPacketizingIncomingMsgs)
        {
            return Task.Run(() =>
            {
                // Run test
                for (int j = 1; j < 5; j++)
                {
                    System.Diagnostics.Debug.WriteLine(string.Format("---------------- Iteration = {0} ---------------", j));
                    // Create client session
                    short clientId = rp1210Api.ClientConnect(client, deviceId, protocol, TxBufferSize, RxBufferSize, isPacketizingIncomingMsgs);
                    System.Diagnostics.Debug.WriteLine(string.Format("ClientConnect clientId = {0}", clientId));

                    // Check for valid client ID
                    if (clientId >= 0 && clientId < 128)
                    {
                        // Set all filter states to pass
                        short result = rp1210Api.SendCommand(3, clientId, new byte[0], 0);

                        // Set echo transmitted messages
                        result = rp1210Api.SendCommand(16, clientId, new byte[] { 1 }, 1);

                        // Send message
                        byte[] message = new byte[] { 8, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9 };
                        for (int i = 0; i < 5; i++)
                        {
                            result = rp1210Api.SendMessage(clientId, message, (short)message.Length, 0, 0);
                            System.Diagnostics.Debug.WriteLine(string.Format("SendMessage returned {0}", result));
                            Thread.Sleep(10);
                        }

                        // Read echo
                        byte[] response = new byte[512];
                        for (int i = 0; i < 5; i++)
                        {
                            result = rp1210Api.ReadMessage(clientId, response, (short)response.Length, 0);
                            if (result > 0)
                            {
                                System.Diagnostics.Debug.WriteLine(string.Format("Read {0} bytes", result));
                                System.Diagnostics.Debug.WriteLine(hexDump(response, result));
                                //prevent thrashing of non-blocked reads
                                Thread.Sleep(1);
                            }
                        }

                        // start asynchronous read thread task
                        readThreadTask = ReadThreadAsync(clientId);

                        // Send message
                        for (int i = 0; i < 5; i++)
                        {
                            result = rp1210Api.SendMessage(clientId, message, (short)message.Length, 0, 0);
                            System.Diagnostics.Debug.WriteLine(string.Format("SendMessage returned {0}", result));
                            Thread.Sleep(10);
                        }

                        // give read thread time to read echo messages before exiting test
                        Thread.Sleep(3000);

                        // End client session
                        result = rp1210Api.ClientDisconnect(clientId);
                        System.Diagnostics.Debug.WriteLine(string.Format("ClientDisconnect returned {0}", result));
                    }
                    else
                    {
                        ShowAlert(string.Format("Connection error: {0}", clientId));
                        break;
                    }
                }

                System.Diagnostics.Debug.WriteLine("---------------- Test Complete ---------------");

                RunOnUiThread(() =>
                {
                    StatusTextView.SetText("Test complete!", TextView.BufferType.Normal);
                });
            });
        }

        public static String hexDump(byte[] ba, int len)
        {
            StringBuilder hex = new StringBuilder(ba.Length * 3);
            for (int i = 0; i < len; i++)
                hex.Append(String.Format("{0:X2}", ba[i]) + " ");

            return hex.ToString();
        }
    }
}

