﻿using Android.Content;
using Rp1210Example;
using System;
using System.Runtime.InteropServices;

namespace Rp1210Example
{
	class Rp1210Api : IRp1210Api
	{
		const int IOCTL_SET_DATA_PATH = 0x102;

		delegate short ClientConnectDelegate(int client, short deviceId, string protocol, int txBufferSize, int rxBufferSize, short isAppPacketizingIncomingMsgs);
		delegate short ClientDisconnectDelegate(short clientId);
		delegate short SendMessageDelegate(short clientId, byte[] clientMessage, short messageSize, short notifyStatusOnTx, short blockOnSend);
		delegate short ReadMessageDelegate(short clientId, byte[] apiMessage, short bufferSize, short blockOnRead);
		delegate short SendCommandDelegate(short commandNumber, short clientId, byte[] clientCommand, short messageSize);
		delegate void ReadVersionDelegate(byte[] dllMajor, byte[] dllMinor, byte[] apiMajor, byte[] apiMinor);
		delegate short GetErrorMsgDelegate(short errorCode, byte[] description);
		delegate short GetHardwareStatusDelegate(short clientId, byte[] clientInfo, short infoSize, short blockOnRequest);
		delegate short ReadDetailedVersionDelegate(short clientId, byte[] apiVersion, byte[] dllVersion, byte[] fwVersion);
		delegate short IoctlDelegate(short clientId, int ioctlId, IntPtr input, IntPtr output);
		delegate void SetDataPath(string path);

		ClientConnectDelegate clientConnectDelegate;
		ClientDisconnectDelegate clientDisconnectDelegate;
		SendMessageDelegate sendMessageDelegate;
		ReadMessageDelegate readMessageDelegate;
		SendCommandDelegate sendCommandDelegate;
		ReadVersionDelegate readVersionDelegate;
		GetErrorMsgDelegate getErrorMsgDelegate;
		GetHardwareStatusDelegate getHardwareStatusDelegate;
		ReadDetailedVersionDelegate readDetailedVersionDelegate;
		IoctlDelegate ioctlDelegate;
		SetDataPath setDataPathDelegate;

		struct Rp1210SdataPath
		{
			public IntPtr dataPath;
		}

		public short ClientConnect(int client, short deviceId, string protocol, int txBufferSize, int rxBufferSize, short isAppPacketizingIncomingMsgs)
		{
			return clientConnectDelegate(client, deviceId, protocol, txBufferSize, rxBufferSize, isAppPacketizingIncomingMsgs);
		}

		public short ClientDisconnect(short clientId)
		{
			return clientDisconnectDelegate(clientId);
		}

		public short SendMessage(short clientId, byte[] clientMessage, short messageSize, short notifyStatusOnTx, short blockOnSend)
		{
			return sendMessageDelegate(clientId, clientMessage, messageSize, notifyStatusOnTx, blockOnSend);
		}

		public short ReadMessage(short clientId, byte[] buffer, short bufferSize, short blockOnRead)
		{
			return readMessageDelegate(clientId, buffer, bufferSize, blockOnRead);
		}

		public short SendCommand(short commandNumber, short clientId, byte[] clientCommand, short messageSize)
		{
			return sendCommandDelegate(commandNumber, clientId, clientCommand, messageSize);
		}

		public void ReadVersion(byte[] dllMajor, byte[] dllMinor, byte[] apiMajor, byte[] apiMinor)
		{
			readVersionDelegate(dllMajor, dllMinor, apiMajor, apiMinor);
		}

		public short GetErrorMsg(short errorCode, byte[] description)
		{
			return getErrorMsgDelegate(errorCode, description);
		}

		public short GetHardwareStatus(short clientId, byte[] clientInfo, short infoSize, short blockOnRequest)
		{
			return getHardwareStatusDelegate(clientId, clientInfo, infoSize, blockOnRequest);
		}

		public short ReadDetailedVersion(short clientId, byte[] apiVersion, byte[] dllVersion, byte[] fwVersion)
		{
			return readDetailedVersionDelegate(clientId, apiVersion, dllVersion, fwVersion);
		}

		public short Ioctl(short clientId, int ioctlId, IntPtr input, IntPtr output)
		{
			return ioctlDelegate(clientId, ioctlId, input, output);
		}

		public void SetImplementation(string implementation)
		{
			if (string.IsNullOrEmpty(implementation))
			{
				throw new ArgumentNullException(nameof(implementation));
			}

			if (string.Compare("NBLR32", implementation, StringComparison.OrdinalIgnoreCase) == 0)
			{
				Java.Lang.JavaSystem.LoadLibrary(implementation.ToLower());
				clientConnectDelegate = nblr32_RP1210_ClientConnect;
				clientDisconnectDelegate = nblr32_RP1210_ClientDisconnect;
				sendMessageDelegate = nblr32_RP1210_SendMessage;
				readMessageDelegate = nblr32_RP1210_ReadMessage;
				sendCommandDelegate = nblr32_RP1210_SendCommand;
				readVersionDelegate = nblr32_RP1210_ReadVersion;
				getErrorMsgDelegate = nblr32_RP1210_GetErrorMsg;
				getHardwareStatusDelegate = nblr32_RP1210_GetHardwareStatus;
				readDetailedVersionDelegate = nblr32_RP1210_ReadDetailedVersion;
				ioctlDelegate = nblr32_RP1210_Ioctl;
				setDataPathDelegate = nblr32_NexiqSetDataPath;
			}
			else if (string.Compare("NULN2R32", implementation, StringComparison.OrdinalIgnoreCase) == 0)
			{
				Java.Lang.JavaSystem.LoadLibrary(implementation.ToLower());
				clientConnectDelegate = nuln2r32_RP1210_ClientConnect;
				clientDisconnectDelegate = nuln2r32_RP1210_ClientDisconnect;
				sendMessageDelegate = nuln2r32_RP1210_SendMessage;
				readMessageDelegate = nuln2r32_RP1210_ReadMessage;
				sendCommandDelegate = nuln2r32_RP1210_SendCommand;
				readVersionDelegate = nuln2r32_RP1210_ReadVersion;
				getErrorMsgDelegate = nuln2r32_RP1210_GetErrorMsg;
				getHardwareStatusDelegate = nuln2r32_RP1210_GetHardwareStatus;
				readDetailedVersionDelegate = nuln2r32_RP1210_ReadDetailedVersion;
				ioctlDelegate = nuln2r32_RP1210_Ioctl;
				setDataPathDelegate = nuln2r32_NexiqSetDataPath;
			}
			else if (string.Compare("CIL7R32", implementation, StringComparison.OrdinalIgnoreCase) == 0)
			{
				Java.Lang.JavaSystem.LoadLibrary(implementation.ToLower());
				clientConnectDelegate = cil7r32_RP1210_ClientConnect;
				clientDisconnectDelegate = cil7r32_RP1210_ClientDisconnect;
				sendMessageDelegate = cil7r32_RP1210_SendMessage;
				readMessageDelegate = cil7r32_RP1210_ReadMessage;
				sendCommandDelegate = cil7r32_RP1210_SendCommand;
				readVersionDelegate = cil7r32_RP1210_ReadVersion;
				getErrorMsgDelegate = cil7r32_RP1210_GetErrorMsg;
				getHardwareStatusDelegate = cil7r32_RP1210_GetHardwareStatus;
				readDetailedVersionDelegate = cil7r32_RP1210_ReadDetailedVersion;
				ioctlDelegate = cil7r32_RP1210_Ioctl;
				setDataPathDelegate = cil7r32_NexiqSetDataPath;
			}
			else if (string.Compare("CIMR32", implementation, StringComparison.OrdinalIgnoreCase) == 0)
			{
				Java.Lang.JavaSystem.LoadLibrary(implementation.ToLower());
				clientConnectDelegate = cimr32_RP1210_ClientConnect;
				clientDisconnectDelegate = cimr32_RP1210_ClientDisconnect;
				sendMessageDelegate = cimr32_RP1210_SendMessage;
				readMessageDelegate = cimr32_RP1210_ReadMessage;
				sendCommandDelegate = cimr32_RP1210_SendCommand;
				readVersionDelegate = cimr32_RP1210_ReadVersion;
				getErrorMsgDelegate = cimr32_RP1210_GetErrorMsg;
				getHardwareStatusDelegate = cimr32_RP1210_GetHardwareStatus;
				readDetailedVersionDelegate = cimr32_RP1210_ReadDetailedVersion;
				ioctlDelegate = cimr32_RP1210_Ioctl;
				setDataPathDelegate = cimr32_NexiqSetDataPath;
			}
			else
			{
				throw new ArgumentException(string.Format("The specified implementation '{0}' isn't valid", implementation));
			}


			var path = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
			IntPtr pathValue = (IntPtr)Marshal.StringToHGlobalAnsi(path);
			Rp1210SdataPath sdataPath = new Rp1210SdataPath()
			{
				dataPath = pathValue
			};

			int sizeDataPath = Marshal.SizeOf(sdataPath);
			IntPtr ptrSetDataPathInput = Marshal.AllocHGlobal(sizeDataPath);
			Marshal.StructureToPtr(sdataPath, ptrSetDataPathInput, false);
			IntPtr ptrSetDataPathOutput = Marshal.AllocHGlobal(1);	// not used...
			short result = ioctlDelegate(0, IOCTL_SET_DATA_PATH, ptrSetDataPathInput, ptrSetDataPathOutput);
			if(result != 0)
			{
				System.Diagnostics.Debug.WriteLine("Set data path failed.");
			}

			Marshal.FreeHGlobal(pathValue);
			Marshal.FreeHGlobal(ptrSetDataPathInput);
			Marshal.FreeHGlobal(ptrSetDataPathOutput);

			//setDataPathDelegate(path);
		}

		#region Cummins Inline 7

		[DllImport("cil7r32")]
		static extern short cil7r32_RP1210_ClientConnect(
			int client,
			short deviceId,
			string protocol,
			int transmitBufferSize,
			int receiveBufferSize,
			short isAppPacketizing);

		[DllImport("cil7r32")]
		static extern short cil7r32_RP1210_ClientDisconnect(
			short clientId);

		[DllImport("cil7r32")]
		static extern short cil7r32_RP1210_GetErrorMsg(
			short errorCode,
			byte[] description);

		[DllImport("cil7r32")]
		static extern short cil7r32_RP1210_GetHardwareStatus(
			short clientId,
			byte[] clientInfo,
			short infoSize,
			short blockOnRequest);

		[DllImport("cil7r32")]
		static extern short cil7r32_RP1210_ReadDetailedVersion(
			short clientId,
			byte[] apiVersion,
			byte[] dllVersion,
			byte[] firmwareVersion);

		[DllImport("cil7r32")]
		static extern short cil7r32_RP1210_ReadMessage(
			short clientId,
			[MarshalAs(UnmanagedType.LPArray), In] byte[] buffer,
			short bufferSize,
			short blockOnRead);

		[DllImport("cil7r32")]
		static extern void cil7r32_RP1210_ReadVersion(
			byte[] dllMajor,
			byte[] dllMinor,
			byte[] apiMajor,
			byte[] apiMinor);

		[DllImport("cil7r32")]
		static extern short cil7r32_RP1210_SendCommand(
			short commandNumber,
			short clientId,
			[MarshalAs(UnmanagedType.LPArray), In] byte[] clientCommand,
			short messageSize);

		[DllImport("cil7r32")]
		static extern short cil7r32_RP1210_SendMessage(
			short clientId,
			[MarshalAs(UnmanagedType.LPArray), In] byte[] message,
			short messageLength,
			short notify,
			short block);

		[DllImport("cil7r32")]
		static extern short cil7r32_RP1210_Ioctl(
			short clientId,
			int ioctlId,
			IntPtr input,
			IntPtr output);

		[DllImport("cil7r32", EntryPoint = "NexiqSetDataPath")]
		static extern void cil7r32_NexiqSetDataPath(
			string path);

		#endregion

		#region Cummins Inline mini

		[DllImport("cimr32")]
		static extern short cimr32_RP1210_ClientConnect(
			int client,
			short deviceId,
			string protocol,
			int transmitBufferSize,
			int receiveBufferSize,
			short isAppPacketizing);

		[DllImport("cimr32")]
		static extern short cimr32_RP1210_ClientDisconnect(
			short clientId);

		[DllImport("cimr32")]
		static extern short cimr32_RP1210_GetErrorMsg(
			short errorCode,
			byte[] description);

		[DllImport("cimr32")]
		static extern short cimr32_RP1210_GetHardwareStatus(
			short clientId,
			byte[] clientInfo,
			short infoSize,
			short blockOnRequest);

		[DllImport("cimr32")]
		static extern short cimr32_RP1210_ReadDetailedVersion(
			short clientId,
			byte[] apiVersion,
			byte[] dllVersion,
			byte[] firmwareVersion);

		[DllImport("cimr32")]
		static extern short cimr32_RP1210_ReadMessage(
			short clientId,
			[MarshalAs(UnmanagedType.LPArray), In] byte[] buffer,
			short bufferSize,
			short blockOnRead);

		[DllImport("cimr32")]
		static extern void cimr32_RP1210_ReadVersion(
			byte[] dllMajor,
			byte[] dllMinor,
			byte[] apiMajor,
			byte[] apiMinor);

		[DllImport("cimr32")]
		static extern short cimr32_RP1210_SendCommand(
			short commandNumber,
			short clientId,
			[MarshalAs(UnmanagedType.LPArray), In] byte[] clientCommand,
			short messageSize);

		[DllImport("cimr32")]
		static extern short cimr32_RP1210_SendMessage(
			short clientId,
			[MarshalAs(UnmanagedType.LPArray), In] byte[] message,
			short messageLength,
			short notify,
			short block);

		[DllImport("cimr32")]
		static extern short cimr32_RP1210_Ioctl(
			short clientId,
			int ioctlId,
			IntPtr input,
			IntPtr output);

		[DllImport("cimr32", EntryPoint = "NexiqSetDataPath")]
		static extern void cimr32_NexiqSetDataPath(
			string path);

		#endregion

		#region Nexiq USB-Link 2

		[DllImport("nuln2r32")]
		static extern short nuln2r32_RP1210_ClientConnect(
			int client,
			short deviceId,
			string protocol,
			int transmitBufferSize,
			int receiveBufferSize,
			short isAppPacketizing);

		[DllImport("nuln2r32")]
		static extern short nuln2r32_RP1210_ClientDisconnect(
			short clientId);

		[DllImport("nuln2r32")]
		static extern short nuln2r32_RP1210_GetErrorMsg(
			short errorCode,
			byte[] description);

		[DllImport("nuln2r32")]
		static extern short nuln2r32_RP1210_GetHardwareStatus(
			short clientId,
			byte[] clientInfo,
			short infoSize,
			short blockOnRequest);

		[DllImport("nuln2r32")]
		static extern short nuln2r32_RP1210_ReadDetailedVersion(
			short clientId,
			byte[] apiVersion,
			byte[] dllVersion,
			byte[] firmwareVersion);

		[DllImport("nuln2r32")]
		static extern short nuln2r32_RP1210_ReadMessage(
			short clientId,
			[MarshalAs(UnmanagedType.LPArray), In] byte[] buffer,
			short bufferSize,
			short blockOnRead);

		[DllImport("nuln2r32")]
		static extern void nuln2r32_RP1210_ReadVersion(
			byte[] dllMajor,
			byte[] dllMinor,
			byte[] apiMajor,
			byte[] apiMinor);

		[DllImport("nuln2r32")]
		static extern short nuln2r32_RP1210_SendCommand(
			short commandNumber,
			short clientId,
			[MarshalAs(UnmanagedType.LPArray), In] byte[] clientCommand,
			short messageSize);

		[DllImport("nuln2r32")]
		static extern short nuln2r32_RP1210_SendMessage(
			short clientId,
			[MarshalAs(UnmanagedType.LPArray), In] byte[] message,
			short messageLength,
			short notify,
			short block);

		[DllImport("nuln2r32")]
		static extern short nuln2r32_RP1210_Ioctl(
			short clientId,
			int ioctlId,
			IntPtr input,
			IntPtr output);

		[DllImport("nuln2r32", EntryPoint = "NexiqSetDataPath")]
		static extern void nuln2r32_NexiqSetDataPath(
			string path);

		#endregion

		#region Nexiq BlueLink

		[DllImport("nblr32")]
		static extern short nblr32_RP1210_ClientConnect(
			int client,
			short deviceId,
			string protocol,
			int transmitBufferSize,
			int receiveBufferSize,
			short isAppPacketizing);

		[DllImport("nblr32")]
		static extern short nblr32_RP1210_ClientDisconnect(
			short clientId);

		[DllImport("nblr32")]
		static extern short nblr32_RP1210_GetErrorMsg(
			short errorCode,
			byte[] description);

		[DllImport("nblr32")]
		static extern short nblr32_RP1210_GetHardwareStatus(
			short clientId,
			byte[] clientInfo,
			short infoSize,
			short blockOnRequest);

		[DllImport("nblr32")]
		static extern short nblr32_RP1210_ReadDetailedVersion(
			short clientId,
			byte[] apiVersion,
			byte[] dllVersion,
			byte[] firmwareVersion);

		[DllImport("nblr32")]
		static extern short nblr32_RP1210_ReadMessage(
			short clientId,
			[MarshalAs(UnmanagedType.LPArray), In] byte[] buffer,
			short bufferSize,
			short blockOnRead);

		[DllImport("nblr32")]
		static extern void nblr32_RP1210_ReadVersion(
			byte[] dllMajor,
			byte[] dllMinor,
			byte[] apiMajor,
			byte[] apiMinor);

		[DllImport("nblr32")]
		static extern short nblr32_RP1210_SendCommand(
			short commandNumber,
			short clientId,
			[MarshalAs(UnmanagedType.LPArray), In] byte[] clientCommand,
			short messageSize);

		[DllImport("nblr32")]
		static extern short nblr32_RP1210_SendMessage(
			short clientId,
			[MarshalAs(UnmanagedType.LPArray), In] byte[] message,
			short messageLength,
			short notify,
			short block);

		[DllImport("nblr32")]
		static extern short nblr32_RP1210_Ioctl(
			short clientId,
			int ioctlId,
			IntPtr input,
			IntPtr output);

		[DllImport("nblr32", EntryPoint = "NexiqSetDataPath")]
		static extern void nblr32_NexiqSetDataPath(
			string path);

		#endregion
	}
}
